# -*- coding: utf-8 -*-

# Copyright (C) 2012-2018 Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#

from __future__ import unicode_literals

import dnf.yum.misc

import tests.support


class TestGenericHolder(tests.support.TestCase):
    def test_merge_lists(self):
        gh = dnf.yum.misc.GenericHolder()
        gh2 = dnf.yum.misc.GenericHolder()
        gh.l = ["lucy", "in", "the"]
        gh2.l = ["sky"]
        gh2.l2 = ["with", "diamonds"]
        gh.merge_lists(gh2)
        self.assertEqual(gh.l, ["lucy", "in", "the", "sky"])
        self.assertEqual(gh.l2, ["with", "diamonds"])
